from collections import namedtuple
from datetime import datetime

from pyramid.authentication import SessionAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy
from pyramid.compat import is_nonstr_iter
from pyramid.httpexceptions import HTTPBadRequest, HTTPServiceUnavailable, HTTPUnauthorized
from pyramid.location import lineage
from pyramid.security import ACLAllowed, ACLDenied, Allow, Authenticated, Everyone, unauthenticated_userid

from parsys_utilities.openid_client import OpenID, request_user_info


Right = namedtuple('Right', ['tenant', 'name'])


def rights_without_tenants(principals):
    """ Transform list of rights in the form (tenant_id, right_name) to list of right_name.

    This is usefull to know if a user has access to an action which does not require tenanting (e.g. list users).
    This function does not remove special Pyramid rights (Authenticated, Everyone).

    Args:
        principals (list of tuples): rights, under the form (tenant_id, right_name).

    Returns:
        set: list of rights names.

    """
    rights = set()
    for principal in principals:
        if isinstance(principal, Right):
            rights.add(principal.name)
        elif isinstance(principal, tuple):
            rights.add(principal[1])
        else:
            rights.add(principal)
    return rights


def get_user(request):
    user_id = unauthenticated_userid(request)
    if not user_id:
        request.session.invalidate()
        return

    # If the user is a service, we don't need to check the access token.
    is_service = request.session.get('user_is_service', False)
    if is_service:
        return request.session['user']

    access_token = request.session.get('access_token')
    # If the access token is still valid, use the user info.
    if access_token and access_token['expires_at'] > datetime.utcnow():
        return request.session['user']

    # If the access token is expired but the refresh token is valid, retrieve a new access token and use the user info.
    refresh_token = request.session.get('refresh_token')
    if refresh_token and refresh_token['expires_at'] > datetime.utcnow():
        # RTA returns the same refresh token, so there is no need to persist it.
        new_access_token, _ = OpenID(request).request_access_token_with_refresh_token(refresh_token['token'])
        request.session['access_token'] = new_access_token
        return request.session['user']

    # If both tokens are expired, logout.
    request.session.invalidate()


def get_user_locale(request):
    if request.user:
        return request.user['locale']


# noinspection PyUnusedLocal
def get_effective_principals(userid, request, allow_admins=False):
    if request.user:
        if allow_admins and request.user['is_admin']:
            return ['g:admin']
        else:
            return [Right(tenant=item[0], name=item[1]) for item in request.user['rights']]


class PatchedSessionAuthenticationPolicy(SessionAuthenticationPolicy):
    def _clean_principal(self, princid):
        """ Remove reserved rights (Authenticated, Everyone, g:admin and userid:...).

        Args:
            princid (str): Right name.

        Returns:
            str: princid or None if right was reserved.

        """
        if princid in (Authenticated, Everyone, 'g:admin') or princid.startswith('userid:'):
            princid = None
        return princid

    def effective_principals(self, request):
        """ Copy/Paste of parent method except for the one line commented. Refer to Pyramid for the full doc.

        As Pyramid adds the user_id to the rights lists, prevent a stupid attack where the user_id would be the name of
        a right, thus giving access to an unauthorized user.

        """
        effective_principals = [Everyone]
        userid = self.unauthenticated_userid(request)

        if userid is None:
            return effective_principals

        if self._clean_principal(userid) is None:
            return effective_principals

        if self.callback is None:
            groups = []
        else:
            groups = self.callback(userid, request)

        if groups is None:
            return effective_principals

        effective_principals.append(Authenticated)
        effective_principals.append('userid:' + userid)
        effective_principals.extend(groups)

        return effective_principals


class OpenIDConnectAuthenticationPolicy(PatchedSessionAuthenticationPolicy):
    def __init__(self, *args, **kwargs):
        self.authorize_services = kwargs.pop('authorize_services', False)
        super().__init__(*args, **kwargs)

    @staticmethod
    def get_user_info_with_access_token(request):
        # If there is no user info in the session but there is a valid access or refresh token
        # (case: returning from settings page), request the user info.
        access_token = request.session.get('access_token')
        if access_token and access_token['expires_at'] > datetime.utcnow():
            rta_userinfo_url = request.route_url('rta', path='api/oauth/userinfo/')
            request.session['user'] = request_user_info(rta_userinfo_url, access_token['token'])

            return request.session['user']['id']

    def get_user_info_with_refresh_token(self, request):
        refresh_token = request.session.get('access_token')
        if refresh_token and refresh_token['expires_at'] > datetime.utcnow():
            new_access_token, refresh_token = OpenID(request).request_access_token_with_refresh_token(['token'])
            request.session['access_token'] = new_access_token

            return self.get_user_info_with_access_token(request)

    @staticmethod
    def get_legacy_access_token(request):
        return request.headers.get('X-Cosmos-Auth-Token') or request.GET.get('X-COSMOS-AUTH-TOKEN')

    @staticmethod
    def get_services_access_token(request):
        access_token = request.headers.get('Authorization')
        if not access_token or not access_token.startswith('Bearer '):
            return
        else:
            return access_token[len('Bearer '):]

    def authenticate_service(self, request):
        access_token = self.get_legacy_access_token(request) or self.get_services_access_token(request)
        if not access_token:
            return

        rta_userinfo_url = request.route_url('rta', path='api/oauth/userinfo/')
        try:
            user_info = request_user_info(rta_userinfo_url, access_token)
        except (HTTPBadRequest, HTTPServiceUnavailable, HTTPUnauthorized):
            return
        else:
            if user_info['is_service']:
                request.session['user'] = user_info
                request.session['user_is_service'] = True
                return user_info['id']
            else:
                return

    def unauthenticated_userid(self, request):
        """ Read user_id from session or manage services.

        Args:
            request (pyramid.request.Request): Request.

        Returns:
            str: userid (None if cookie does not exist/is invalid).

        """
        try:
            return request.session['user']['id']
        except (KeyError, TypeError):
            pass

        user_id = self.get_user_info_with_access_token(request)
        if user_id:
            return user_id

        user_id = self.get_user_info_with_refresh_token(request)
        if user_id:
            return user_id

        if self.authorize_services:
            return self.authenticate_service(request)


class TenantedAuthorizationPolicy(ACLAuthorizationPolicy):
    """ Tenanted Pyramid ACLAuthorizationPolicy. """
    def permits(self, context, principals, permission):
        """ Copy/Paste of parent method except for the 2 lines commented. Refer to Pyramid for the full doc.

        Basically instead of using the classic Pyramid model of ACL (Allow, principal, permission), we use
        (Allow, tenant, principal, permission). With one catch: some actions do not require tenant (e.g. list objects)
        and the 'admin' principal renders tenanting useless.
        """
        acl = '<No ACL found on any object in resource lineage>'

        for location in lineage(context):
            try:
                acl = location.__acl__
            except AttributeError:
                continue

            if acl and callable(acl):
                acl = acl()

            for ace in acl:
                # 4 variables instead of 3.
                ace_action, ace_tenant, ace_principal, ace_permissions = ace
                # If the ACL is not tenanted, only check the right name, otherwise compare the rights with the
                # (tenant_id, right_name) format.
                if (ace_tenant is None and ace_principal in rights_without_tenants(principals)) or \
                        ((ace_tenant, ace_principal) in principals):
                    if not is_nonstr_iter(ace_permissions):
                        ace_permissions = [ace_permissions]
                    if permission in ace_permissions:
                        if ace_action == Allow:
                            return ACLAllowed(ace, acl, permission,
                                              principals, location)
                        else:
                            return ACLDenied(ace, acl, permission,
                                             principals, location)

        return ACLDenied('<default deny>', acl, permission, principals, context)
