from datetime import datetime, timedelta
from uuid import uuid4
from urllib.parse import parse_qs, quote, unquote, urlencode, urlparse

from jwt import decode, ExpiredSignatureError, InvalidAudienceError, InvalidIssuerError
from oauthlib.oauth2 import OAuth2Error, WebApplicationClient
from pyramid.httpexceptions import HTTPBadRequest, HTTPFound, HTTPUnauthorized, HTTPServiceUnavailable
from pyramid.renderers import render_to_response
from pyramid.security import Allow, Authenticated
from pyramid.settings import asbool
from pyramid.view import forbidden_view_config, view_config
from requests import ConnectionError, Timeout, get, post

from parsys_utilities.logging import utilities_logger


def validate_response(response):
    if response.status_code == 200:
        return response.json()
    elif response.status_code == 400:
        raise HTTPBadRequest()
    elif response.status_code in {401, 403}:
        raise HTTPUnauthorized()
    else:
        raise HTTPServiceUnavailable()


def request_user_info(url, access_token):
    try:
        response = get(url, headers={'authorization': 'Bearer ' + access_token})
    except (ConnectionError, Timeout):
        raise HTTPServiceUnavailable()

    return validate_response(response)


class OpenID(object):
    def __init__(self, request):
        self.request = request
        self.openid_client_id = self.request.registry.settings['rta.client_id']
        self.openid_client_secret = self.request.registry.settings['rta.secret']
        self.openid_client = WebApplicationClient(self.openid_client_id)

    def verify_authorization_response(self, url, session_csrf_token):
        try:
            # Retrieve information from Oauth authorization success.
            authorization_response = self.openid_client.parse_request_uri_response(url)
            authorization_code = authorization_response['code']
            state_encoded = authorization_response['state']
            state = parse_qs(unquote(state_encoded))

            csrf_token = state['csrf_token'][0]
            user_target_path = state['target_path'][0]
            user_target_path = urlparse(user_target_path).path
        except (IndexError, KeyError, TypeError):
            raise HTTPBadRequest()

        if csrf_token != session_csrf_token:
            raise HTTPBadRequest()

        return authorization_code, user_target_path

    def verify_json_web_token(self, jwt, issuer):
        now = datetime.utcnow()

        try:
            signed_id_token = jwt['id_token']
            access_token_expires_at = now + timedelta(seconds=int(jwt['access_token_expires_in']))
            access_token = {'token': jwt['access_token'], 'expires_at': access_token_expires_at}
            refresh_token_expires_at = now + timedelta(seconds=int(jwt['refresh_token_expires_in']))
            refresh_token = {'token': jwt['refresh_token'], 'expires_at': refresh_token_expires_at}
        except (KeyError, TypeError):
            raise HTTPBadRequest()

        try:
            # Even if we don't persist it, make sure that the id_token is valid.
            decode(signed_id_token, self.openid_client_secret, audience=self.openid_client_id, issuer=issuer)
        except (ExpiredSignatureError, InvalidAudienceError, InvalidIssuerError, OAuth2Error):
            raise HTTPBadRequest()
        else:
            return access_token, refresh_token

    def request_access_token_with_authorization_code(self, url, authorization_code, client_redirect_endpoint):
        # Using authorization code, request token.
        data = self.openid_client.prepare_request_body(code=authorization_code, redirect_uri=client_redirect_endpoint)
        try:
            response = post(url, data=data, auth=(self.openid_client_id, self.openid_client_secret))
        except (ConnectionError, Timeout):
            raise HTTPServiceUnavailable()

        if response.status_code == 400:
            raise HTTPBadRequest()
        elif response.status_code != 200:
            raise HTTPServiceUnavailable()

        return response.json()

    def request_access_token_with_refresh_token(self, refresh_token):
        refresh_data = self.openid_client.prepare_refresh_body(refresh_token=refresh_token)

        rta_token_url = self.request.route_url('rta', path='api/oauth/token/')
        try:
            response = post(rta_token_url, data=refresh_data, auth=(self.openid_client_id, self.openid_client_secret))
        except (ConnectionError, Timeout):
            raise HTTPServiceUnavailable()

        json_web_token = validate_response(response)
        issuer = self.request.registry.settings['rta.server_url']
        return self.verify_json_web_token(json_web_token, issuer)

    @forbidden_view_config(renderer='errors/500.html')
    def manage_forbidden(self):
        if self.request.user:
            self.request.response.status_int = 403
            return render_to_response('errors/403.html', {}, request=self.request, response=self.request.response)

        is_pjax_request = asbool(self.request.headers.get('X-Pjax'))
        if self.request.is_xhr or is_pjax_request or not self.request.accept.accept_html():
            return HTTPUnauthorized()

        request_token_endpoint = self.request.route_url('oauth-request_token')
        rta_authorization_url = self.request.route_url('rta', path='api/oauth/authorize/')
        csrf_token = self.request.session.new_csrf_token()
        state = {'csrf_token': csrf_token, 'target_path': self.request.path}
        encoded_state = quote(urlencode(state))

        rta_authorization_request = \
            self.openid_client.prepare_request_uri(rta_authorization_url, scope=['openid'],
                                                   redirect_uri=request_token_endpoint, state=encoded_state)

        response = HTTPFound(location=rta_authorization_request)
        return response

    @view_config(route_name='oauth-request_token', request_method='GET')
    def request_token(self):
        csrf_token = self.request.session.get('_csrft_')
        if not csrf_token:
            self.request.response.status_int = 403
            return HTTPFound(location=self.request.route_url('home'))

        authorization_code, user_target_path = self.verify_authorization_response(self.request.url, csrf_token)

        rta_token_url = self.request.route_url('rta', path='api/oauth/token/')
        request_token_endpoint = self.request.route_url('oauth-request_token')
        json_web_token = self.request_access_token_with_authorization_code(rta_token_url, authorization_code,
                                                                           request_token_endpoint)

        issuer = self.request.registry.settings['rta.server_url']
        access_token, refresh_token = self.verify_json_web_token(json_web_token, issuer)

        rta_userinfo_url = self.request.route_url('rta', path='api/oauth/userinfo/')
        user_info = request_user_info(rta_userinfo_url, access_token['token'])

        self.request.session['user'] = user_info
        self.request.session['session_id'] = uuid4().hex
        self.request.session['access_token'] = access_token
        self.request.session['refresh_token'] = refresh_token

        utilities_logger(self.request).info('log in')

        return HTTPFound(location=user_target_path)


class Users(object):
    __acl__ = [
        (Allow, None, Authenticated, 'authenticated')
    ]

    def __init__(self, request):
        self.request = request

    @view_config(route_name='users-logout', request_method='GET')
    def logout_get(self):
        self.request.session.invalidate()

        client_id = self.request.registry.settings['rta.client_id']
        rta_logout_url = self.request.route_url('rta', path='users/logout/', _query={'redirect': client_id})

        utilities_logger(self.request).info('log out')

        return HTTPFound(location=rta_logout_url)

    @view_config(route_name='users-settings', request_method='GET', permission='authenticated')
    def settings_get(self):
        self.request.session.pop('user')
        return HTTPFound(location=self.request.route_url('rta', path='users/settings/'))


def includeme(config):
    config.add_route(pattern='users/logout/', name='users-logout')
    config.add_route(pattern='users/settings/', name='users-settings', factory=Users)
    config.add_route(pattern='oauth/request_token/', name='oauth-request_token')
