from re import findall

from pyramid.i18n import TranslationString as _
from pyramid.settings import asbool
from sqlalchemy.exc import InvalidRequestError

from parsys_utilities.model import asc, Boolean, desc, func, or_, String


def get_count(q):
    count_q = q.statement.with_only_columns([func.count()]).order_by(None)
    return q.session.execute(count_q).scalar()


def sql_search(db_session, searched_object, full_text_search_attributes, joined_tables=None, tenanting=None,
               specific_search_attributes=None, specific_sort_attributes=None, search_parameters=None):
    """ Perform a standard SQL query on our DB.

    To cater to all special needs, the infos of the queried object can be specified (full text search, special sorts,
    etc.).
    Search parameters:
        full_text_search (str).
        search (list of tuples): Tuples are (object attribute, search value).
        sort (list of str): Each element should be an attribute of the object or defined in the specific sort
        attributes.
        Sorting is done in the order of the elements in the list. Precede the element by '-' to do a descending sort.

    Args:
        db_session (sqlalchemy.session.Session): Database session.
        searched_object (sqlalchemy.schema.Table): Objects type to search.
        full_text_search_attributes (list of sqlalchemy.Columns): Object's attributes to search on during full text
        search.
        joined_tables (list of sqlalchemy.schema.Table): Tables to join on (for full text search/tenanting/search).
        tenanting (function sqlalchemy.query.query:sqlalchemy.query.query): Function which must be applied to return
        tenanted list.
        specific_search_attributes (dict): SQLAlchemy column to search on when the search term doesn't match the
        corresponding object's attribute.
        specific_sort_attributes (dict): SQLAlchemy column to sort on when the sort term doesn't match the
        corresponding object's attribute.
        search_parameters (dict): Limit/offset/search term/sort/full text search.

    Returns:
        dict:
                items (int): Searched objects.
                recordsTotal (int): Total number of objects after tenanting.
                recordsFiltered (int): Number of filtered objects (because of search/full text search).

    """
    search_parameters = search_parameters or {}
    specific_search_attributes = specific_search_attributes or []
    specific_sort_attributes = specific_sort_attributes or []
    output = {}

    q = db_session.query(searched_object)
    for table in joined_tables or []:
        q = q.outerjoin(table)

    # Apply tenanting first. Users should only know the number of objects for their tenant.
    if tenanting:
        q = tenanting(q)

    output.update({'recordsTotal': get_count(q)})

    # Sort list.
    for sort_element in search_parameters.get('sort') or []:
        if sort_element.startswith('-'):
            sort_function = desc
            sort_element = sort_element.lstrip('-')
        else:
            sort_function = asc

        if sort_element in specific_sort_attributes:
            q = q.order_by(sort_function(specific_sort_attributes[sort_element]))

        elif hasattr(searched_object, sort_element):
            attribute = getattr(searched_object, sort_element)
            try:
                # When sorting strings, always try to do a case-insensitive sort.
                if isinstance(attribute.type, String):
                    q = q.order_by(sort_function(func.lower(attribute)))
                else:
                    q = q.order_by(sort_function(attribute))
            # Hybrid properties will fail here, so do a simple sort.
            except (AttributeError, InvalidRequestError):
                q = q.order_by(sort_function(attribute))

        # If we didn't find the attribute to sort on, fail.
        else:
            raise KeyError(_('Unknown attribute: {}.').format(sort_element))

    # Filter list.
    for search_element in search_parameters.get('search') or []:
        if search_element[0] in specific_search_attributes:
            attribute = specific_search_attributes[search_element[0]]

        elif hasattr(searched_object, search_element[0]):
            attribute = getattr(searched_object, search_element[0])

        else:
            raise KeyError(_('Unknown attribute: {}.').format(search_element[0]))

        search_variable = search_element[2]
        try:
            # Transform pseudo Booleans (like 1 or 'True') into real Booleans.
            if search_variable in ['false', 'true', 'False', 'True'] or isinstance(attribute.type, Boolean):
                search_variable = asbool(search_element[2])
        except (AttributeError, InvalidRequestError):
            pass

        if search_element[1] == '==':
            q = q.filter(attribute == search_variable)
        elif search_element[1] == '!=':
            q = q.filter(attribute != search_variable)

    # Apply full text search.
    if search_parameters.get('full_text_search'):
        words = findall(r'\w+', search_parameters['full_text_search'])
        for word in words:
            sql_filter = (attribute.ilike('%{}%'.format(word)) for attribute in full_text_search_attributes)
            q = q.filter(or_(sql_filter))

    # Number of records after filters.
    output.update({'recordsFiltered': get_count(q)})

    if search_parameters.get('limit'):
        q = q.limit(search_parameters['limit'])

    if search_parameters.get('offset'):
        q = q.offset(search_parameters['offset'])

    results = q.all()

    output.update({'items': results})
    return output
